"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.handler = exports.commandNames = exports.help = exports.cliOptionsTypes = exports.rcOptionsTypes = void 0;
const path_1 = __importDefault(require("path"));
const cli_utils_1 = require("@pnpm/cli-utils");
const constants_1 = require("@pnpm/constants");
const error_1 = require("@pnpm/error");
const read_project_manifest_1 = require("@pnpm/read-project-manifest");
const store_connection_manager_1 = require("@pnpm/store-connection-manager");
const graceful_fs_1 = __importDefault(require("@pnpm/graceful-fs"));
const core_1 = require("@pnpm/core");
const config_1 = require("@pnpm/config");
const workspace_find_packages_1 = require("@pnpm/workspace.find-packages");
const logger_1 = require("@pnpm/logger");
const sort_packages_1 = require("@pnpm/sort-packages");
const rimraf_1 = __importDefault(require("@zkochan/rimraf"));
const load_json_file_1 = __importDefault(require("load-json-file"));
const map_1 = __importDefault(require("ramda/src/map"));
const render_help_1 = __importDefault(require("render-help"));
const lockfile_1 = require("@yarnpkg/lockfile");
const yarnCore = __importStar(require("@yarnpkg/core"));
const parsers_1 = require("@yarnpkg/parsers");
const path_exists_1 = __importDefault(require("path-exists"));
const recursive_1 = require("../recursive");
const yarnUtil_1 = require("./yarnUtil");
var YarnLockType;
(function (YarnLockType) {
    YarnLockType["yarn"] = "yarn";
    YarnLockType["yarn2"] = "yarn2";
})(YarnLockType || (YarnLockType = {}));
exports.rcOptionsTypes = cliOptionsTypes;
function cliOptionsTypes() {
    return {};
}
exports.cliOptionsTypes = cliOptionsTypes;
function help() {
    return (0, render_help_1.default)({
        description: `Generates ${constants_1.WANTED_LOCKFILE} from an npm package-lock.json (or npm-shrinkwrap.json, yarn.lock) file.`,
        url: (0, cli_utils_1.docsUrl)('import'),
        usages: [
            'pnpm import',
        ],
    });
}
exports.help = help;
exports.commandNames = ['import'];
async function handler(opts, params) {
    // Removing existing pnpm lockfile
    // it should not influence the new one
    await (0, rimraf_1.default)(path_1.default.join(opts.dir, constants_1.WANTED_LOCKFILE));
    const versionsByPackageNames = {};
    let preferredVersions = {};
    if (await (0, path_exists_1.default)(path_1.default.join(opts.dir, 'yarn.lock'))) {
        const yarnPackageLockFile = await readYarnLockFile(opts.dir);
        getAllVersionsFromYarnLockFile(yarnPackageLockFile, versionsByPackageNames);
    }
    else if (await (0, path_exists_1.default)(path_1.default.join(opts.dir, 'package-lock.json')) ||
        await (0, path_exists_1.default)(path_1.default.join(opts.dir, 'npm-shrinkwrap.json'))) {
        const npmPackageLock = await readNpmLockfile(opts.dir);
        if (npmPackageLock.lockfileVersion < 3) {
            getAllVersionsByPackageNamesPreV3(npmPackageLock, versionsByPackageNames);
        }
        else {
            getAllVersionsByPackageNames(npmPackageLock, versionsByPackageNames);
        }
    }
    else {
        throw new error_1.PnpmError('LOCKFILE_NOT_FOUND', 'No lockfile found');
    }
    preferredVersions = getPreferredVersions(versionsByPackageNames);
    // For a workspace with shared lockfile
    if (opts.workspaceDir) {
        const allProjects = opts.allProjects ?? await (0, workspace_find_packages_1.findWorkspacePackages)(opts.workspaceDir, opts);
        const selectedProjectsGraph = opts.selectedProjectsGraph ?? selectProjectByDir(allProjects, opts.dir);
        if (selectedProjectsGraph != null) {
            const sequencedGraph = (0, sort_packages_1.sequenceGraph)(selectedProjectsGraph);
            // Check and warn if there are cyclic dependencies
            if (!opts.ignoreWorkspaceCycles && !sequencedGraph.safe) {
                const cyclicDependenciesInfo = sequencedGraph.cycles.length > 0
                    ? `: ${sequencedGraph.cycles.map(deps => deps.join(', ')).join('; ')}`
                    : '';
                if (opts.disallowWorkspaceCycles) {
                    throw new error_1.PnpmError('DISALLOW_WORKSPACE_CYCLES', `There are cyclic workspace dependencies${cyclicDependenciesInfo}`);
                }
                logger_1.logger.warn({
                    message: `There are cyclic workspace dependencies${cyclicDependenciesInfo}`,
                    prefix: opts.workspaceDir,
                });
            }
            await (0, recursive_1.recursive)(allProjects, params, 
            // @ts-expect-error
            {
                ...opts,
                lockfileOnly: true,
                selectedProjectsGraph,
                preferredVersions,
                workspaceDir: opts.workspaceDir,
            }, 'import');
        }
        return;
    }
    const store = await (0, store_connection_manager_1.createOrConnectStoreController)(opts);
    const manifest = await (0, read_project_manifest_1.readProjectManifestOnly)(opts.dir);
    const manifestOpts = opts.rootProjectManifest ? (0, config_1.getOptionsFromRootManifest)(opts.rootProjectManifestDir, opts.rootProjectManifest) : {};
    const installOpts = {
        ...opts,
        ...manifestOpts,
        lockfileOnly: true,
        preferredVersions,
        storeController: store.ctrl,
        storeDir: store.dir,
    };
    await (0, core_1.install)(manifest, installOpts);
}
exports.handler = handler;
async function readYarnLockFile(dir) {
    try {
        const yarnLockFile = await graceful_fs_1.default.readFile(path_1.default.join(dir, 'yarn.lock'), 'utf8');
        const yarnLockFileType = getYarnLockfileType(yarnLockFile);
        if (yarnLockFileType === YarnLockType.yarn) {
            const lockJsonFile = (0, lockfile_1.parse)(yarnLockFile);
            if (lockJsonFile.type === 'success') {
                return lockJsonFile.object;
            }
            else {
                throw new error_1.PnpmError('YARN_LOCKFILE_PARSE_FAILED', `Yarn.lock file was ${lockJsonFile.type}`);
            }
        }
        else if (yarnLockFileType === YarnLockType.yarn2) {
            const lockJsonFile = parseYarn2Lock(yarnLockFile);
            if (lockJsonFile.type === YarnLockType.yarn2) {
                return lockJsonFile.object;
            }
        }
    }
    catch (err) { // eslint-disable-line
        if (err['code'] !== 'ENOENT')
            throw err;
    }
    throw new error_1.PnpmError('YARN_LOCKFILE_NOT_FOUND', 'No yarn.lock found');
}
function parseYarn2Lock(lockFileContents) {
    const parseYarnLock = (0, parsers_1.parseSyml)(lockFileContents);
    delete parseYarnLock.__metadata;
    const dependencies = {};
    const { structUtils } = yarnCore;
    const { parseDescriptor, parseRange } = structUtils;
    const keyNormalizer = (0, yarnUtil_1.yarnLockFileKeyNormalizer)(parseDescriptor, parseRange);
    Object.entries(parseYarnLock).forEach(
    // eslint-disable-next-line
    ([fullDescriptor, versionData]) => {
        keyNormalizer(fullDescriptor).forEach((descriptor) => {
            dependencies[descriptor] = versionData;
        });
    });
    return {
        object: dependencies,
        type: YarnLockType.yarn2,
    };
}
async function readNpmLockfile(dir) {
    try {
        return await (0, load_json_file_1.default)(path_1.default.join(dir, 'package-lock.json'));
    }
    catch (err) { // eslint-disable-line
        if (err['code'] !== 'ENOENT')
            throw err;
    }
    try {
        return await (0, load_json_file_1.default)(path_1.default.join(dir, 'npm-shrinkwrap.json'));
    }
    catch (err) { // eslint-disable-line
        if (err['code'] !== 'ENOENT')
            throw err;
    }
    throw new error_1.PnpmError('NPM_LOCKFILE_NOT_FOUND', 'No package-lock.json or npm-shrinkwrap.json found');
}
function getPreferredVersions(versionsByPackageNames) {
    const preferredVersions = (0, map_1.default)((versions) => Object.fromEntries(Array.from(versions).map((version) => [version, 'version'])), versionsByPackageNames);
    return preferredVersions;
}
function getAllVersionsByPackageNamesPreV3(npmPackageLock, versionsByPackageNames) {
    if (npmPackageLock.dependencies == null)
        return;
    for (const [packageName, { version }] of Object.entries(npmPackageLock.dependencies)) {
        if (!versionsByPackageNames[packageName]) {
            versionsByPackageNames[packageName] = new Set();
        }
        versionsByPackageNames[packageName].add(version);
    }
    for (const dep of Object.values(npmPackageLock.dependencies)) {
        getAllVersionsByPackageNamesPreV3(dep, versionsByPackageNames);
    }
}
function getAllVersionsByPackageNames(pkg, versionsByPackageNames) {
    if (pkg.dependencies) {
        extractDependencies(versionsByPackageNames, pkg.dependencies);
    }
    if ('packages' in pkg && pkg.packages) {
        extractDependencies(versionsByPackageNames, pkg.packages);
    }
}
function extractDependencies(versionsByPackageNames, dependencies) {
    for (let [pkgName, pkgDetails] of Object.entries(dependencies)) {
        if (pkgName.includes('node_modules')) {
            pkgName = pkgName.substring(pkgName.lastIndexOf('node_modules/') + 13);
        }
        if (!versionsByPackageNames[pkgName]) {
            versionsByPackageNames[pkgName] = new Set();
        }
        if (pkgDetails.version) {
            versionsByPackageNames[pkgName].add(pkgDetails.version);
        }
        if (pkgDetails.packages) {
            extractDependencies(versionsByPackageNames, pkgDetails.packages);
        }
        if (pkgDetails.dependencies) {
            for (const [pkgName1, version] of Object.entries(pkgDetails.dependencies)) {
                if (!versionsByPackageNames[pkgName1]) {
                    versionsByPackageNames[pkgName1] = new Set();
                }
                versionsByPackageNames[pkgName1].add(version);
            }
        }
    }
}
function getAllVersionsFromYarnLockFile(yarnPackageLock, versionsByPackageNames) {
    for (const [packageName, { version }] of Object.entries(yarnPackageLock)) {
        const pkgName = packageName.substring(0, packageName.lastIndexOf('@'));
        if (!versionsByPackageNames[pkgName]) {
            versionsByPackageNames[pkgName] = new Set();
        }
        versionsByPackageNames[pkgName].add(version);
    }
}
function selectProjectByDir(projects, searchedDir) {
    const project = projects.find(({ dir }) => path_1.default.relative(dir, searchedDir) === '');
    if (project == null)
        return undefined;
    return { [searchedDir]: { dependencies: [], package: project } };
}
function getYarnLockfileType(lockFileContents) {
    return lockFileContents.includes('__metadata')
        ? YarnLockType.yarn2
        : YarnLockType.yarn;
}
//# sourceMappingURL=index.js.map